import math
from typing import List, Tuple, Optional, Dict


class NavmeshNode:
    '''Class for representing one node in the navigation mesh. Each node is a polygon with some additional data:
        -) group index - the index of the connected component of polygons
        -) polygon index
        -) array of indexes of incident polygons
        -) indexes of polygon vertices
        -) coordinates of the polygon center
        -) portals (this is a map, where key - index of the neighbor node, value - the pair of 3-tuples ((v1), (v2)), where v1 and v2 coordinates of the edge vertices )
                    ordered in orientetion, induced from polygon orientation

        polygons are convex, it's important for calculation it normal

        this class is for internal use only
    '''
    def __init__(self, all_vertices: List[Tuple[float, float, float]], index: int, polygon_indexes: List[int]):
        '''Pass to the constructor all vertices of the navigation mesh
        '''
        self._vertices: List[Tuple[float, float, float]] = [all_vertices[i] for i in polygon_indexes]
        self._polygon: List[int] = polygon_indexes  # indexes of polygon vertices
        self._group: int = -1
        self._index: int = index
        self._neighbors: List[int] = []
        self._center: Tuple[float, float, float] = self._calc_center(self._vertices)
        self._normal: Tuple[float, float, float] = self._calc_average_normal(self._center, self._vertices)
        self._portals: Dict[int, Tuple[Tuple[float, float, float], Tuple[float, float, float]]] = {}
        self._vertex_normals: List[Tuple[float, float, float]] = self._calc_vertex_normals(self._vertices)

    def add_neighbor(self, node_index: int, vertex_0: Tuple[float, float, float], vertex_1: Tuple[float, float, float]):
        '''add neighbor polygon and portal
        '''
        if node_index not in self._neighbors:
            self._neighbors.append(node_index)
            self._portals[node_index] = (vertex_0, vertex_1)
        else:
            print("[Something wrong] Try to add neighbor node " + str(node_index) + " to the " + str(self._index) + ", but this node already exists")

    def get_portal(self, node_index: int) -> Tuple[Tuple[float, float, float], Tuple[float, float, float]]:
        '''return pair of 3-tuples with coordinates of the edge, which separate current node from node with index node_index
        '''
        return self._portals[node_index]

    def get_neighbord(self) -> List[int]:
        return self._neighbors

    def get_index(self) -> int:
        return self._index

    def get_group(self) -> int:
        return self._group

    def get_vertex_indexes(self) -> List[int]:
        return self._polygon

    def get_vertex_coordinates(self) -> List[Tuple[float, float, float]]:
        return self._vertices

    def get_normal(self) -> Tuple[float, float, float]:
        return self._normal

    def get_center(self) -> Tuple[float, float, float]:
        return self._center

    def set_group(self, group_index: int, group_array: List[int], all_nodes):  # all_nodes: List[NavmeshNode]
        if self._group == -1:
            self._group = group_index
            group_array.append(self._index)
            for n in self._neighbors:
                all_nodes[n].set_group(group_index, group_array, all_nodes)

    def is_point_inside(self, point: Tuple[float, float, float]) -> bool:
        '''return true, if the point inside the polygon
        '''
        size: int = len(self._vertices)
        for i in range(size):
            u: Tuple[float, float, float] = self._vertices[i]
            v: Tuple[float, float, float] = self._vertices[self._modulize(i + 1, size)]
            # we should calculate cross product [uv, up]
            # if dot-product with normal in the vertex < 0, then point outside the polygon
            vector: Tuple[float, float, float] = self._cross(v[0] - u[0], v[1] - u[1], v[2] - u[2], point[0] - u[0], point[1] - u[1], point[2] - u[2])
            d: float = vector[0] * self._vertex_normals[i][0] + vector[1] * self._vertex_normals[i][1] + vector[2] * self._vertex_normals[i][2]
            if d < 0.0:
                return False
        return True

    def _modulize(self, value: int, n: int) -> int:
        if value < n:
            return value
        else:
            return value - n

    def _calc_vertex_normals(self, vertices: List[Tuple[float, float, float]]) -> List[Tuple[float, float, float]]:
        '''for each vertex u (and next v, w) calculate normalized [uv, uw]
        '''
        size: int = len(vertices)
        to_return: List[Tuple[float, float, float]] = []
        for i in range(size):
            u: Tuple[float, float, float] = vertices[i]
            v: Tuple[float, float, float] = vertices[self._modulize(i + 1, size)]
            w: Tuple[float, float, float] = vertices[self._modulize(i + 2, size)]
            vector: Tuple[float, float, float] = self._cross(v[0] - u[0], v[1] - u[1], v[2] - u[2], w[0] - u[0], w[1] - u[1], w[2] - u[2])
            d: float = math.sqrt(vector[0]**2 + vector[1]**2 + vector[2]**2)
            to_return.append((vector[0] / d, vector[1] / d, vector[2] / d))
        return to_return

    def _calc_center(self, array: List[Tuple[float, float, float]]) -> Tuple[float, float, float]:
        '''Find center of the array
        '''
        s = [0.0] * 3
        for a in array:
            for i in range(len(a)):
                s[i] += a[i]
        for i in range(len(s)):
            s[i] = s[i] / len(array)
        return (s[0], s[1], s[2])

    def _cross(self, a_x: float, a_y: float, a_z: float, b_x: float, b_y: float, b_z: float) -> Tuple[float, float, float]:
        return (a_y*b_z - a_z*b_y, a_z*b_x - a_x*b_z, a_x*b_y - a_y*b_x)

    def _calc_average_normal(self, center: Tuple[float, float, float], points: List[Tuple[float, float, float]]) -> Tuple[float, float, float]:
        normal: List[float] = [0.0, 0.0, 0.0]
        for p_index in range(len(points)):
            p0: Tuple[float, float, float] = points[p_index]
            p1: Tuple[float, float, float] = points[p_index + 1 if p_index < len(points) - 1 else 0]
            c: Tuple[float, float, float] = self._cross(p0[0] - center[0], p0[1] - center[1], p0[2] - center[2], p1[0] - center[0], p1[1] - center[1], p1[2] - center[2])
            normal[0] += c[0]
            normal[1] += c[1]
            normal[2] += c[2]
        d: float = math.sqrt(normal[0]**2 + normal[1]**2 + normal[2]**2)
        return (normal[0] / d, normal[1] / d, normal[2] / d)

    def __repr__(self):
        return "<node " + str(self._index) +\
               ", vertices: " + str(self._polygon) +\
               ", normal: " + str(self._normal) +\
               ", local normals: " + str(self._vertex_normals) +\
               ", neigh: " + str(self._neighbors) +\
               ", component: " + str(self._group) +\
               ", portals: " + str(self._portals) +\
               ">"



from typing import List, Tuple, Optional
from navmesh.navmesh_graph import NavmeshGraph
from navmesh.navmesh_node import NavmeshNode
from navmesh.navmesh_bvh import NavmeshBVH


class Navmesh:
    '''Main class for pathfinding process
    '''
    def __init__(self, vertices: List[Tuple[float, float, float]], polygons: List[List[int]]):
        '''Init pathfinder object by setting vertices and polygons of the navmesh

        Input:
            vertices - array of the form [(x1, y1, z1), (x2, y2, z1), ...]
            polygons - array of the form [p1, p2, ...], where each pi is array [i1, i2, ..., in], ij - indexes of vertices
                       the sequence [i1, i2, ..., in] should be in clockwise direction from the main point of view
                       this is importatnt for search path algorithm
                       if polygon orientation will be conter-clockwise, then the path will be incorrect

        Example of the simple square grid with two 4-sided polygons
            vertices = [(1.0, 0.0, 1.0), (-1.0, 0.0, 1.0), (-1.0, 0.0, -1.0), (1.0, 0.0, -1.0), (0.0, 0.0, 1.0), (0.0, 0.0, -1.0)]
            polygons = [[0, 3, 5, 4], [4, 5, 2, 1]]
        '''
        self._vertices: List[Tuple[float, float, float]] = vertices
        self._polygons: List[List[int]] = polygons
        self._graphs: List[NavmeshGraph] = []  # graph, where vertices are centers of polygons, edges are pairs of two incident (by edge only!) polygons
        self._nodes: List[NavmeshNode] = []
        self._groups: List[List[int]] = []  # each group is an array of node indexes with the same group index

        vertex_map: List[List[int]] = [[] for v in range(len(vertices))]
        for p_index in range(len(polygons)):
            # iterate throw polygons
            for p_v_index in polygons[p_index]:
                # add polygon index to the array of corresponding vertices
                vertex_map[p_v_index].append(p_index)
            # create the node
            self._nodes.append(NavmeshNode(vertices, p_index, polygons[p_index]))

        # next we can use vertex_map for define neneighbors of each polygon
        for node in self._nodes:
            node_vertices: List[int] = node.get_vertex_indexes()
            # for each vertex of the node (polygon) we should get indexes of incident polygons and take intersections of each sequential pair
            for v_num in range(len(node_vertices)):
                u: int = node_vertices[v_num]
                v: int = node_vertices[v_num + 1 if v_num < len(node_vertices) - 1 else 0]
                intersection: List[int] = self._get_intersection(vertex_map[u], vertex_map[v])
                if len(intersection) == 0:
                    print("[Someting wrong] Intersection of polygons, incident to vertices " + str(u) + " and " + str(v) + " are empty")
                elif len(intersection) == 2:
                    # in principle, other cases are impossible
                    node_index: int = node.get_index()
                    if node_index not in intersection:
                        print("[Something wrong] Polygon " + str(node_index) + " does not contains in the neighborhood of tow incident vertices")
                    else:
                        for i in intersection:
                            if i != node_index:
                                node.add_neighbor(i, self._vertices[u], self._vertices[v])
                elif len(intersection) > 2:
                    print("[Someting wrong] Intersection of polygons, incident to vertices " + str(u) + " and " + str(v) + " contains " + str(len(intersection)) + " items " + str(intersection))

        # define groups
        for node in self._nodes:
            g: int = node.get_group()
            if g  == -1:
                # if we get hte first polygon with undefined group
                new_group: List[int] = []  # start new group array
                new_index: int = len(self._groups)  # generate the next group index
                node.set_group(new_index, new_group, self._nodes)  # start recursive provess
                self._groups.append(new_group)

        # finally, define the graph
        # one graph for each group
        for group in self._groups:
            graph_edges: List[Tuple[int, int]] = []  # each graph is an aray of pairs (n1, n2), where n1 and n2 are node (=polygon) indexes and n1 < n2
            for node_index in group:
                node_neighbors: List[int] = self._nodes[node_index].get_neighbord()
                for other_node in node_neighbors:
                    edge: Tuple[int, int] = (node_index, other_node) if node_index < other_node else (other_node, node_index)
                    # if this edge is new, add it to the graph
                    if edge not in graph_edges:
                        graph_edges.append(edge)
            # next we should get all graph node indexes, sort it and get it centers
            graph_verts: List[int] = []
            for edge in graph_edges:
                for i in range(2):
                    if edge[i] not in graph_verts:
                        graph_verts.append(edge[i])
            graph_verts.sort()
            self._graphs.append(NavmeshGraph([self._nodes[i].get_center() for i in graph_verts], graph_verts, graph_edges))

        # build bvh
        self._bvh: NavmeshBVH = NavmeshBVH(self._nodes)

    def search_path(self, start: Tuple[float, float, float], finish: Tuple[float, float, float]) -> List[Tuple[float, float, float]]:
        '''Search the path between start and finish points in the navigation mesh

        Input:
            start - 3-tuple (x, y, z) of the start point
            finish - 3-tuple (x, y, z) of the finish point

        Return:
            array in the form [(x1, y1, z1), (x2, y2, z2), ...] with coordinates of points, which form the output path
            start and finish points include as the first and the last entries in the array
            if there is no path between input points, then return empty array []
        '''
        # find nodes indexes for start and end point
        start_node: Optional[NavmeshNode] = self._bvh.sample(start)
        finish_node: Optional[NavmeshNode] = self._bvh.sample(finish)
        if start_node is not None and finish_node is not None:
            # check are nodes in one group
            start_index: int = start_node.get_index()
            finish_index: int = finish_node.get_index()
            group_index: int = self._get_nodes_group_index(start_index, finish_index)
            if group_index > -1:
                graph: NavmeshGraph = self._graphs[group_index]
                # find path between nodes in the graph
                graph_path: List[int] = graph.search(start_index, finish_index)

                # next create non-optimal path throw portals
                raw_path: List[Tuple[float, float, float]] = [start, start]
                for p_i in range(1, len(graph_path)):
                    # extend raw path by portal points between p_i-th node and p_i+1-th
                    portal: Tuple[Tuple[float, float, float], Tuple[float, float, float]] = self._nodes[graph_path[p_i - 1]].get_portal(graph_path[p_i])
                    raw_path.extend(portal)
                raw_path.extend([finish, finish])

                # finally, simplify the raw_path, by using pull the rope algorithm
                # get it from https://github.com/donmccurdy/three-pathfinding
                portal_apex: Tuple[float, float, float] = raw_path[0]
                portal_left: Tuple[float, float, float] = raw_path[0]
                portal_right: Tuple[float, float, float] = raw_path[1]

                apex_index: int = 0
                left_index: int = 0
                right_index: int = 0

                finall_path: List[Tuple[float, float, float]] = [portal_apex]
                i: int = 1
                while i < len(raw_path) // 2:
                    left: Tuple[float, float, float] = raw_path[2 * i]
                    right: Tuple[float, float, float] = raw_path[2 * i + 1]

                    skip_next: bool = False
                    # update right vertex
                    if self._triangle_area_2(portal_apex, portal_right, right) <= 0.0:
                        if self._v_equal(portal_apex, portal_right) or self._triangle_area_2(portal_apex, portal_left, right) > 0.0:
                            portal_right = right
                            right_index = i
                        else:
                            finall_path.append(portal_left)
                            # make current left the new apex
                            portal_apex = portal_left
                            apex_index = left_index
                            # reset portal
                            portal_left = portal_apex
                            portal_right = portal_apex
                            left_index = apex_index
                            right_index = apex_index
                            # restart scan
                            i = apex_index
                            skip_next = True
                    if not skip_next:
                        # update left vertex
                        if self._triangle_area_2(portal_apex, portal_left, left) >= 0.0:
                            if self._v_equal(portal_apex, portal_left) or self._triangle_area_2(portal_apex, portal_right, left) < 0.0:
                                portal_left = left
                                left_index = i
                            else:
                                finall_path.append(portal_right)
                                # make current right the new apex
                                portal_apex = portal_right
                                apex_index = right_index
                                # reset portal
                                portal_left = portal_apex
                                portal_right = portal_apex
                                left_index = apex_index
                                right_index = apex_index
                                # restart scan
                                i = apex_index
                    i += 1
                if (len(finall_path) == 0 or not self._v_equal(finall_path[len(finall_path) - 1], raw_path[len(raw_path) - 2])):
                    # append last point to path
                    finall_path.append(raw_path[len(raw_path) - 2])
                return finall_path
            else:
                # nodes in the different groups, so, there are no path between them
                return []
        else:
            # start or finish node is None, so, no path
            return []

    def _v_equal(self, a: Tuple[float, float, float], b: Tuple[float, float, float], epsilon: float = 0.01) -> bool:
        '''a, b are points

        return True, if distance from a to b less, then epsilon
        '''
        return (a[0] - b[0])**2 + (a[1] - b[1])**2 + (a[2] - b[2])**2 < epsilon

    def _triangle_area_2(self, a: Tuple[float, float, float], b: Tuple[float, float, float], c: Tuple[float, float, float]) -> float:
        '''a, b, and c are 3-tuples with vertex coordinates
        '''
        return (c[0] - a[0]) * (b[2] - a[2]) - (b[0] - a[0]) * (c[2] - a[2])

    def _check_nodes_in_one_group(self, index_01: int, index_02: int) -> bool:
        for group in self._groups:
            if index_01 in group and index_02 in group:
                return True
        return False

    def _get_nodes_group_index(self, index_01: int, index_02: int) -> int:
        '''return index of the group with indexes index_01 and index_02, -1 if these values are in different groups
        '''
        for i in range(len(self._groups)):
            group: List[int] = self._groups[i]
            if index_01 in group and index_02 in group:
                return i
        return -1

    def _get_intersection(self, array_a: List[int], array_b: List[int]) -> List[int]:
        '''return intersection of two arrays
        '''
        to_return: List[int] = []
        for a in array_a:
            if a in array_b:
                to_return.append(a)
        return to_return



from typing import List, Tuple, Optional
from navmesh.navmesh_node import NavmeshNode

BVH_AABB_DELTA = 0.5


class NavmeshBVH:
    '''Class for the node int bvh-tree
    '''
    def __init__(self, nodes: List[NavmeshNode]):
        '''Create one node (and call recursive building all children nodes) from array of NavmeshNodes

        Input:
            nodes - array of NavmeshNode objects
        '''
        self._node: Optional[NavmeshNode] = None
        self._left: Optional[NavmeshBVH] = None
        self._right: Optional[NavmeshBVH] = None

        x_min: float = float("inf")
        x_max: float = -float("inf")
        y_min: float = float("inf")
        y_max: float = -float("inf")
        z_min: float = float("inf")
        z_max: float = -float("inf")
        self._aabb: Tuple[float, float, float, float, float, float]

        if len(nodes) == 1:
            self._node = nodes[0]
            # build aabb
            verts: List[Tuple[float, float, float]] = self._node.get_vertex_coordinates()
            for v in verts:
                if v[0] < x_min:
                    x_min = v[0]
                if v[0] > x_max:
                    x_max = v[0]
                if v[1] < y_min:
                    y_min = v[1]
                if v[1] > y_max:
                    y_max = v[1]
                if v[2] < z_min:
                    z_min = v[2]
                if v[2] > z_max:
                    z_max = v[2]
            self._aabb = (x_min - BVH_AABB_DELTA, y_min - BVH_AABB_DELTA, z_min - BVH_AABB_DELTA, x_max + BVH_AABB_DELTA, y_max + BVH_AABB_DELTA, z_max + BVH_AABB_DELTA)
        else:
            # find the axis (x or z) to split the space
            x_median: float = 0.0
            z_median: float = 0.0
            for node in nodes:
                c: Tuple[float, float, float] = node.get_center()
                x_median += c[0]
                z_median += c[2]
                if c[0] < x_min:
                    x_min = c[0]
                if c[0] > x_max:
                    x_max = c[0]
                if c[2] < z_min:
                    z_min = c[2]
                if c[2] > z_max:
                    z_max = c[2]
            split_axis: int = 0 if (x_max - x_min) > (z_max - z_min) else 2
            median: float = x_median / len(nodes) if (x_max - x_min) > (z_max - z_min) else z_median / len(nodes)
            left: List[NavmeshNode] = []
            right: List[NavmeshNode] = []
            for node in nodes:
                if node.get_center()[split_axis] < median:
                    left.append(node)
                else:
                    right.append(node)
            if len(left) == 0:
                # move last right node to the left array
                left.append(right.pop())
            else:
                # left array is not empty, but may be empty right array
                if len(right) == 0:
                    right.append(left.pop())
            self._left = NavmeshBVH(left)
            self._right = NavmeshBVH(right)
            l_aabb: Tuple[float, float, float, float, float, float] = self._left.get_aabb()
            r_aabb: Tuple[float, float, float, float, float, float] = self._right.get_aabb()
            self._aabb = self._union_aabbs(l_aabb, r_aabb)

    def _union_aabbs(self, b1: Tuple[float, float, float, float, float, float], b2: Tuple[float, float, float, float, float, float]) -> Tuple[float, float, float, float, float, float]:
        return (min(b1[0], b2[0]), min(b1[1], b2[1]), min(b1[2], b2[2]),
                max(b1[3], b2[3]), max(b1[4], b2[4]), max(b1[5], b2[5]))

    def get_aabb(self) -> Tuple[float, float, float, float, float, float]:
        '''Return 6-tuple of the aabb (axis align bounding box) of the current node

        this 6-tuple is (x_min, y_min, z_min, x_max, y_max, z_max)
        '''
        return self._aabb

    def is_inside_aabb(self, point: Tuple[float, float, float]) -> bool:
        '''Return True, if the point is inside the current aabb, and False otherwise

        Input:
            point - 3-triple (x, y, z)

        Output:
            True or False
        '''
        return self._aabb[0] < point[0] and self._aabb[1] < point[1] and self._aabb[2] < point[2] and\
               self._aabb[3] > point[0] and self._aabb[4] > point[1] and self._aabb[5] > point[2]

    def sample(self, point: Tuple[float, float, float]) -> Optional[NavmeshNode]:
        '''Return node, which contains the point
        If there are no nodes near the point, then return None

        Input:
            point - 3-triple (x, y, z)
        
        Output:
            NavmeshNode or None
        '''
        if self.is_inside_aabb(point):
            if self._node is None and self._left is not None and self._right is not None:
                # this bvh node contains left and right children, go deeper
                left_sample: Optional[NavmeshNode] = self._left.sample(point)
                right_sample: Optional[NavmeshNode] = self._right.sample(point)
                if left_sample is None:
                    return right_sample
                else:
                    if right_sample is None:
                        return left_sample
                    else:
                        # we should choose from left and right sample the closest to the point
                        l_c: Tuple[float, float, float] = left_sample.get_center()
                        l_n: Tuple[float, float, float] = left_sample.get_normal()
                        l_dist: float = abs((point[0] - l_c[0]) * l_n[0] + (point[1] - l_c[1]) * l_n[1] + (point[2] - l_c[2]) * l_n[2])

                        r_c: Tuple[float, float, float] = right_sample.get_center()
                        r_n: Tuple[float, float, float] = right_sample.get_normal()
                        r_dist: float = abs((point[0] - r_c[0]) * r_n[0] + (point[1] - r_c[1]) * r_n[1] + (point[2] - r_c[2]) * r_n[2])

                        if l_dist < r_dist:
                            return left_sample
                        else:
                            return right_sample
            else:
                # bvh node contains polygon
                if self._node is not None:
                    if self._node.is_point_inside(point):
                        return self._node
                    else:
                        return None
                else:
                    return None
        else:
            return None

    def __repr__(self) -> str:
        return "<object: " + str(self._node if self._node is None else self._node.get_index()) + ", left: " + str(self._left) + ", right: " + str(self._right) + ">"



import math
from typing import List, Tuple, Dict


class NavmeshGraph:
    '''Graph with specific data, needed for A* algorithm
    '''
    def __init__(self, vertex_positions: List[Tuple[float, float, float]], vertices: List[int], edges: List[Tuple[int, int]]):
        '''Inicialize the graph from it vertices and edges

        Input:
            vertex_positions - array of 3-tuples [(x, y, z), ...]
            vertices - array of integer names of the vertices
            edges - array of 2-tuple [(a, b), ...], where a and b are names of vertices, incident to the same edge
        '''
        self._positions: List[Tuple[float, float, float]] = vertex_positions
        self._vertex_names: List[int] = vertices  # names are not from 0 to n-1
        self._vertex_count: int = len(self._vertex_names)
        # we need oppisit map from vertex name to it index
        self._index_map: Dict[int, int] = {}
        for i in range(len(self._vertex_names)):
            self._index_map[self._vertex_names[i]] = i
        self._edges: List[Tuple[int, int]] = edges

        # create map from vertex to all incident vertices
        self._incident_map: Dict[int, List[int]] = {}  # key - vertex index, value - array of indexes of incident vertices
        for e in self._edges:
            for i in range(2):
                v = self._index_map[e[i]]
                if v in self._incident_map:
                    self._incident_map[v].append(self._index_map[e[i-1]])
                else:
                    self._incident_map[v] = [self._index_map[e[i-1]]]

    def _pre_start(self, target: int):
        self._vertices_g: List[float] = [0.0] * self._vertex_count
        self._vertices_h: List[float] = [self._get_distance(self._positions[i], self._positions[target]) for i in range(self._vertex_count)]  # calculate distance from each vertex to the target
        self._vertices_f: List[float] = [0.0] * self._vertex_count
        self._vertices_close: List[bool] = [False] * self._vertex_count
        self._vertices_parent: List[int] = [-1] * self._vertex_count  # store here the index of the vertex, from we come here with minimum g

    def _get_distance(self, a: Tuple[float, float, float], b: Tuple[float, float, float]) -> float:
        return math.sqrt((a[0] - b[0])**2 + (a[1] - b[1])**2 + (a[2] - b[2])**2)

    def get_vertex_count(self):
        return self._vertex_count

    def get_vertices(self):
        '''Return graph vertex names
        '''
        return self._vertex_names

    def get_positions(self):
        '''Return the array of graph vertex positions
        '''
        return self._positions

    def get_edges(self):
        '''Return array of graph edges
        '''
        return self._edges

    def search(self, start_vertex: int, end_vertex: int) -> List[int]:
        '''Calculate the shortes path between two input vertices, by using A* algorithm

        Input:
            start_vertex - the name of the start vertex
            end_vertex - the name of the end vertex

        Output:
            array of vertex names, which form the shortest path between start end end vertices
            if there are no path between input vertices, then return empty array []
        '''
        # clear data structures
        start: int = self._index_map[start_vertex]
        end: int = self._index_map[end_vertex]
        self._pre_start(end)
        # create open list
        open_list: List[int] = [start]
        # set f and g for the start_vertex
        self._vertices_g[start] = 0.0
        self._vertices_f[start] = self._vertices_g[start] + self._vertices_h[start]

        while len(open_list) > 0:
            # find in the open list vertex with the minimum value of f
            min_f: float = self._vertices_f[open_list[0]]
            min_vertex: int = open_list[0]
            min_index: int = 0
            v: int
            for i in range(1, len(open_list)):
                v = open_list[i]
                v_f: float = self._vertices_f[v]
                if v_f < min_f:
                    min_f = v_f
                    min_vertex = v
                    min_index = i
            open_list.pop(min_index)  # remove vertex with minimal f from the open list
            self._vertices_close[min_vertex] = True  # and set it close
            if min_vertex == end:
                # we find the closest path
                path: List[int] = []  # put to the path indexes of vertices
                v = min_vertex
                while self._vertices_parent[v] > -1:
                    path.append(v)
                    v = self._vertices_parent[v]
                path.append(v)
                # return array of vertices names
                l: int = len(path)
                return [self._vertex_names[path[l - 1 - i]] for i in range(l)]
            # next we should enumerate all vertex edges and add all non-initial vertices to the open list
            if min_vertex in self._incident_map:
                for child_index in self._incident_map[min_vertex]:
                    if self._vertices_close[child_index] is False:
                        v_g: float = self._get_distance(self._positions[min_vertex], self._positions[child_index]) + self._vertices_g[min_vertex]
                        if self._vertices_parent[child_index]  == -1:
                            # we come to this vertex at first time, so, set parent and g-value
                            self._vertices_parent[child_index] = min_vertex
                            self._vertices_g[child_index] = v_g
                            self._vertices_f[child_index] = v_g + self._vertices_h[child_index]
                            # add it to the open list
                            open_list.append(child_index)
                        else:
                            # we already come to this vertex, so, this vertex in the open list
                            # check, may be g-value is less, than existing one
                            if self._vertices_g[child_index] > v_g:
                                # redefine parent
                                self._vertices_parent[child_index] = min_vertex
                                self._vertices_g[child_index] = v_g
                                self._vertices_f[child_index] = v_g + self._vertices_h[child_index]
        # empty array mean, that there are no path between vertices in the graph
        return []

    def __repr__(self):
        return "<graph " + str(self._vertex_names) + ", edges: " + str(self._edges) + ", map: " + str(self._index_map) + ", positions: " + str(self._positions) + ">"
